(* metacirc-cont-esc.sml
 *
 * Meta-circular interpreter with explicit continuations and escape.
 *)
structure MetaCircContEsc = struct

    structure OL = OLEsc

    type var = OL.var

    datatype value = INTv of int
		   | BOOLv of bool
		   | FUNv of value * cont -> value

    withtype cont = value -> value

    fun vFUN  (FUNv f)  = f | vFUN _  = raise Fail "function expected"
    fun vBOOL (BOOLv b) = b | vBOOL _ = raise Fail "boolean expected"
    fun vINT  (INTv i)  = i | vINT _  = raise Fail "integer expected"

    type env = var -> value

    fun evcon (OL.INT i) = INTv i
      | evcon (OL.BOOL b) = BOOLv b

    fun eval (r, e, k) =
	case r of
	    OL.CONST c => k (evcon c)
	  | OL.VAR v => k (e v)
	  | OL.APPL { opr, opnd } =>
	      eval (opr, e, fn f => eval (opnd, e, fn a => vFUN f (a, k)))
	  | OL.LAMBDA l => k (evlambda (l, e))
	  | OL.COND { prem, conc, altr } =>
	      eval (prem, e, fn p =>
		if vBOOL p then eval (conc, e, k) else eval (altr, e, k))
	  | OL.LETREC { dvar, dexp, body } =>
	      let fun e' x = if x = dvar then evlambda (dexp, e') else e x
	      in eval (body, e', k)
	      end
	  | OL.ESCAPE { evar, body } =>
	      eval (body, ext (evar, FUNv (fn (a, _) => k a), e), k)

    and evlambda ({ fp, body }, e) =
	FUNv (fn (a, k) => eval (body, ext (fp, a, e), k))

    and ext (z, a, e) x = if x = z then a else e x

    fun equal (INTv i, INTv j) = BOOLv (i = j)
      | equal (BOOLv x, BOOLv y) = BOOLv (x = y)
      | equal _ = raise Fail "illegal argument to equal"
    fun succ x = INTv (vINT x + 1)

    fun initenv "succ" = FUNv (fn (x, k) => k (succ x))
      | initenv "equal" =
	  FUNv (fn (x, k) => k (FUNv (fn (y, k) => k (equal (x, y)))))
      | initenv x = raise Fail ("unbound variable " ^ x)

    fun interpret r = eval (r, initenv, fn a => a)
end
