(* metacirc-defunct-funval.sml
 *
 * The meta-circular interpreter with function values defunctionalized.
 * (Environments are still implemented as functions.)
 *)
structure MetaCircDefunctFunval = struct

    type var = OL.var

    datatype value = INTv of int
		   | BOOLv of bool
		   | FUNv of funval

    and funval = CLOSR of { lam: OL.lambda, en: env }
               | SC
	       | EQ1
	       | EQ2 of value

    withtype env = var -> value

    fun vFUN  (FUNv f) = f  | vFUN _  = raise Fail "function expected"
    fun vBOOL (BOOLv b) = b | vBOOL _ = raise Fail "boolean expected"
    fun vINT  (INTv i) = i  | vINT _  = raise Fail "integer expected"

    fun evcon (OL.INT i) = INTv i
      | evcon (OL.BOOL b) = BOOLv b

    fun apply (CLOSR { lam = { fp, body }, en }, a) =
	  eval (body, ext (fp, a, en))
      | apply (SC, a) = INTv (vINT a + 1)
      | apply (EQ1, a) = FUNv (EQ2 a)
      | apply (EQ2 (INTv i), INTv j) = BOOLv (i = j)
      | apply (EQ2 (BOOLv x), BOOLv y) = BOOLv (x = y)
      | apply (EQ2 _, _) = raise Fail "illegal arguments to EQUAL"

    and eval (r, e) =
	case r of
	    OL.CONST c => evcon c
	  | OL.VAR v => e v
	  | OL.APPL { opr, opnd } =>
	      apply (vFUN (eval (opr, e)), eval (opnd, e))
	  | OL.LAMBDA l => evlambda (l, e)
	  | OL.COND { prem, conc, altr } =>
	      if vBOOL (eval (prem, e)) then eval (conc, e) else eval (altr, e)
	  | OL.LETREC { dvar, dexp, body } =>
	      let fun e' x = if x = dvar then evlambda (dexp, e') else e x
	      in eval (body, e')
	      end

    and evlambda (lam, e) = FUNv (CLOSR { lam = lam, en = e })

    and ext (z, a, e) x = if x = z then a else e x

    fun initenv "succ" = FUNv SC
      | initenv "equal" = FUNv EQ1
      | initenv x = raise Fail ("unbound variable " ^ x)

    fun interpret r = eval (r, initenv)
end
