(* metacirc.sml
 *
 *   The meta-circular interpreter: every language construct is
 *   implemented using the corresponding construct of the meta language.
 *)
structure MetaCirc = struct

    type var = OL.var

    datatype value = INTv of int
		   | BOOLv of bool
		   | FUNv of value -> value

    fun vFUN  (FUNv f)  = f | vFUN _  = raise Fail "function expected"
    fun vBOOL (BOOLv b) = b | vBOOL _ = raise Fail "boolean expected"
    fun vINT  (INTv i)  = i | vINT _  = raise Fail "integer expected"

    type env = var -> value

    fun evcon (OL.INT i) = INTv i
      | evcon (OL.BOOL b) = BOOLv b

    fun eval (r, e) =
	case r of
	    OL.CONST c => evcon c
	  | OL.VAR v => e v
	  | OL.APPL { opr, opnd } => vFUN (eval (opr, e)) (eval (opnd, e))
	  | OL.LAMBDA l => evlambda (l, e)
	  | OL.COND { prem, conc, altr } =>
	      if vBOOL (eval (prem, e)) then eval (conc, e) else eval (altr, e)
	  | OL.LETREC { dvar, dexp, body } =>
	      let fun e' x = if x = dvar then evlambda (dexp, e') else e x
	      in eval (body, e')
	      end

    and evlambda ({ fp, body }, e) =
	FUNv (fn a => eval (body, ext (fp, a, e)))

    and ext (z, a, e) x = if x = z then a else e x

    fun equal (INTv i, INTv j) = BOOLv (i = j)
      | equal (BOOLv x, BOOLv y) = BOOLv (x = y)
      | equal _ = raise Fail "illegal argument to equal"

    fun initenv "succ" = FUNv (fn x => INTv (vINT x + 1))
      | initenv "equal" = FUNv (fn x => FUNv (fn y => equal (x, y)))
      | initenv x = raise Fail ("unbound variable " ^ x)

    fun interpret r = eval (r, initenv)
end
